/***************************************************************************
 * sdla_remora.c	WANPIPE(tm) Multiprotocol WAN Link Driver. 
 *				AFT REMORA and FXO/FXS support module.
 *
 * Author: 	Alex Feldman   <al.feldman@sangoma.com>
 *
 * Copyright:	(c) 2005 Sangoma Technologies Inc.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 * ============================================================================
 * Oct 6, 2005	Alex Feldman	Initial version.
 ******************************************************************************
 */

/*******************************************************************************
**			   INCLUDE FILES
*******************************************************************************/
#if (defined __FreeBSD__) | (defined __OpenBSD__)
# include <net/wanpipe_includes.h>
# include <net/wanpipe_debug.h>
# include <net/wanpipe_defines.h>
# include <net/wanpipe_abstr.h>
# include <net/wanpipe_common.h>
# include <net/wanpipe.h>
# include <net/sdla_remora.h>
#elif (defined __WINDOWS__)
# include <wanpipe\csu_dsu.h>
#else
# include <linux/wanpipe_includes.h>
# include <linux/wanpipe_defines.h>
# include <linux/wanpipe.h>
# include <linux/sdla_remora.h>
#endif

/*******************************************************************************
**			  DEFINES AND MACROS
*******************************************************************************/
//#define	SPI2STEP

#define MOD_CHAIN_DISABLED	0
#define MOD_CHAIN_ENABLED	1

#define SPI_INTERFACE_REG	0x54

#define MOD_SPI_ADDR_FXS_READ	0x80
#define MOD_SPI_ADDR_FXS_WRITE	0x00

#define MOD_SPI_CS_FXS_CHIP_0	0x01
#define MOD_SPI_CS_FXS_CHIP_1	0x02

#define MOD_SPI_CS_FXO_CHIP_1	0x08
#define MOD_SPI_CS_FXO_WRITE	0x00
#define MOD_SPI_CS_FXO_READ	0x40

#define MOD_SPI_CTRL_START	0x08	
#define MOD_SPI_CTRL_FXS	0x10
#define MOD_SPI_CTRL_CHAIN	0x20

#define MOD_SPI_RESET		0x40000000
#define MOD_SPI_BUSY		0x80000000
#define MOD_SPI_START		0x08000000

static alpha  indirect_regs[] =
{
{0,255,"DTMF_ROW_0_PEAK",0x55C2},
{1,255,"DTMF_ROW_1_PEAK",0x51E6},
{2,255,"DTMF_ROW2_PEAK",0x4B85},
{3,255,"DTMF_ROW3_PEAK",0x4937},
{4,255,"DTMF_COL1_PEAK",0x3333},
{5,255,"DTMF_FWD_TWIST",0x0202},
{6,255,"DTMF_RVS_TWIST",0x0202},
{7,255,"DTMF_ROW_RATIO_TRES",0x0198},
{8,255,"DTMF_COL_RATIO_TRES",0x0198},
{9,255,"DTMF_ROW_2ND_ARM",0x0611},
{10,255,"DTMF_COL_2ND_ARM",0x0202},
{11,255,"DTMF_PWR_MIN_TRES",0x00E5},
{12,255,"DTMF_OT_LIM_TRES",0x0A1C},
{13,0,"OSC1_COEF",0x7B30},
{14,1,"OSC1X",0x0063},
{15,2,"OSC1Y",0x0000},
{16,3,"OSC2_COEF",0x7870},
{17,4,"OSC2X",0x007D},
{18,5,"OSC2Y",0x0000},
{19,6,"RING_V_OFF",0x0000},
{20,7,"RING_OSC",0x7EF0},
{21,8,"RING_X",0x0160},
{22,9,"RING_Y",0x0000},
{23,255,"PULSE_ENVEL",0x2000},
{24,255,"PULSE_X",0x2000},
{25,255,"PULSE_Y",0x0000},
//{26,13,"RECV_DIGITAL_GAIN",0x4000},	// playback volume set lower
{26,13,"RECV_DIGITAL_GAIN",0x2000},	// playback volume set lower
{27,14,"XMIT_DIGITAL_GAIN",0x4000},
//{27,14,"XMIT_DIGITAL_GAIN",0x2000},
{28,15,"LOOP_CLOSE_TRES",0x1000},
{29,16,"RING_TRIP_TRES",0x3600},
{30,17,"COMMON_MIN_TRES",0x1000},
{31,18,"COMMON_MAX_TRES",0x0200},
{32,19,"PWR_ALARM_Q1Q2",0x07C0},
{33,20,"PWR_ALARM_Q3Q4",0x2600},
{34,21,"PWR_ALARM_Q5Q6",0x1B80},
{35,22,"LOOP_CLOSURE_FILTER",0x8000},
{36,23,"RING_TRIP_FILTER",0x0320},
{37,24,"TERM_LP_POLE_Q1Q2",0x008C},
{38,25,"TERM_LP_POLE_Q3Q4",0x0100},
{39,26,"TERM_LP_POLE_Q5Q6",0x0010},
{40,27,"CM_BIAS_RINGING",0x0C00},
{41,64,"DCDC_MIN_V",0x0C00},
{42,255,"DCDC_XTRA",0x1000},
{43,66,"LOOP_CLOSE_TRES_LOW",0x1000},
};

static struct fxo_mode {
	char *name;
	/* FXO */
	int ohs;
	int ohs2;
	int rz;
	int rt;
	int ilim;
	int dcv;
	int mini;
	int acim;
	int ring_osc;
	int ring_x;
} fxo_modes[] =
{
	{ "FCC", 0, 0, 0, 1, 0, 0x3, 0, 0, }, 	/* US, Canada */
	{ "TBR21", 0, 0, 0, 0, 1, 0x3, 0, 0x2, 0x7e6c, 0x023a, },
		/* Austria, Belgium, Denmark, Finland, France, Germany, 
		   Greece, Iceland, Ireland, Italy, Luxembourg, Netherlands,
		   Norway, Portugal, Spain, Sweden, Switzerland, and UK */
	{ "ARGENTINA", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "AUSTRALIA", 1, 0, 0, 0, 0, 0, 0x3, 0x3, },
	{ "AUSTRIA", 0, 1, 0, 0, 1, 0x3, 0, 0x3, },
	{ "BAHRAIN", 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "BELGIUM", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "BRAZIL", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "BULGARIA", 0, 0, 0, 0, 1, 0x3, 0x0, 0x3, },
	{ "CANADA", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "CHILE", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "CHINA", 0, 0, 0, 0, 0, 0, 0x3, 0xf, },
	{ "COLUMBIA", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "CROATIA", 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "CYRPUS", 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "CZECH", 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "DENMARK", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "ECUADOR", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "EGYPT", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "ELSALVADOR", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "FINLAND", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "FRANCE", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "GERMANY", 0, 1, 0, 0, 1, 0x3, 0, 0x3, },
	{ "GREECE", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "GUAM", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "HONGKONG", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "HUNGARY", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "ICELAND", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "INDIA", 0, 0, 0, 0, 0, 0x3, 0, 0x4, },
	{ "INDONESIA", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "IRELAND", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "ISRAEL", 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "ITALY", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "JAPAN", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "JORDAN", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "KAZAKHSTAN", 0, 0, 0, 0, 0, 0x3, 0, },
	{ "KUWAIT", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "LATVIA", 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "LEBANON", 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "LUXEMBOURG", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "MACAO", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "MALAYSIA", 0, 0, 0, 0, 0, 0, 0x3, 0, },	/* Current loop >= 20ma */
	{ "MALTA", 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "MEXICO", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "MOROCCO", 0, 0, 0, 0, 1, 0x3, 0, 0x2, },
	{ "NETHERLANDS", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "NEWZEALAND", 0, 0, 0, 0, 0, 0x3, 0, 0x4, },
	{ "NIGERIA", 0, 0, 0, 0, 0x1, 0x3, 0, 0x2, },
	{ "NORWAY", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "OMAN", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "PAKISTAN", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "PERU", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "PHILIPPINES", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "POLAND", 0, 0, 1, 1, 0, 0x3, 0, 0, },
	{ "PORTUGAL", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "ROMANIA", 0, 0, 0, 0, 0, 3, 0, 0, },
	{ "RUSSIA", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "SAUDIARABIA", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "SINGAPORE", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "SLOVAKIA", 0, 0, 0, 0, 0, 0x3, 0, 0x3, },
	{ "SLOVENIA", 0, 0, 0, 0, 0, 0x3, 0, 0x2, },
	{ "SOUTHAFRICA", 1, 0, 1, 0, 0, 0x3, 0, 0x3, },
	{ "SOUTHKOREA", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "SPAIN", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "SWEDEN", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "SWITZERLAND", 0, 1, 0, 0, 1, 0x3, 0, 0x2, },
	{ "SYRIA", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "TAIWAN", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "THAILAND", 0, 0, 0, 0, 0, 0, 0x3, 0, },
	{ "UAE", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "UK", 0, 1, 0, 0, 1, 0x3, 0, 0x5, },
	{ "USA", 0, 0, 0, 0, 0, 0x3, 0, 0, },
	{ "YEMEN", 0, 0, 0, 0, 0, 0x3, 0, 0, },
};

/* Interrupt flag enable */
#if 0
# define WAN_REMORA_INTR
#endif
#if 0
# define WAN_REMORA_FXS_LCIP
#endif
#if 0
# define WAN_REMORA_FXS_DTMF
#endif

/*******************************************************************************
**			STRUCTURES AND TYPEDEFS
*******************************************************************************/

/*******************************************************************************
**			   GLOBAL VARIABLES
*******************************************************************************/
extern WAN_LIST_HEAD(, wan_tdmv_) wan_tdmv_head;
/*******************************************************************************
**			  FUNCTION PROTOTYPES
*******************************************************************************/
static int wp_remora_intr(sdla_fe_t *); 
static int wp_remora_check_intr(sdla_fe_t *); 

/*******************************************************************************
**			  FUNCTION DEFINITIONS
*******************************************************************************/

static void wait_just_a_bit(int foo)
{
	unsigned long	ticks;

	ticks = SYSTEM_TICKS + foo;
	while(SYSTEM_TICKS < ticks);
}

static void wp_remora_reset_spi(sdla_fe_t *fe)
{
	sdla_t	*card = (sdla_t*)fe->card;

	WAN_ASSERT1(card == NULL);
	card->hw_iface.bus_write_4(	card->hw,
					SPI_INTERFACE_REG,
					MOD_SPI_RESET);
	WP_DELAY(1000);
	card->hw_iface.bus_write_4(	card->hw,
					SPI_INTERFACE_REG,
					0x00000000);
	WP_DELAY(1000);
	return;
}

static int
__wp_remora_setreg(sdla_fe_t *fe, int mod_no, int type, int chain, u8 reg, u8 value)
{
	sdla_t		*card = (sdla_t*)fe->card;
	unsigned int	data = 0;
	unsigned char	chip_select = 0x00;
	unsigned char	ctrl_byte =0x00;
	int		i;

	WAN_ASSERT(card == NULL);
	WAN_ASSERT(card->hw_iface.bus_write_4 == NULL);
	WAN_ASSERT(card->hw_iface.bus_read_4 == NULL);

	/* bit 0-7: data byte */
	data = value & 0xFF;
	if (type == MOD_TYPE_FXO){

		/* bit 8-15: register number */
		data |= (reg & 0xFF) << 8;

		/* bit 16-23: chip select byte
		** bit 16
		**
		**
		**			*/
		chip_select = 0x20;
		chip_select |= MOD_SPI_CS_FXO_WRITE;
		if (mod_no % 2 == 0){
			/* Select second chip in a chain */
			chip_select |= MOD_SPI_CS_FXO_CHIP_1;
		}
		data |= (chip_select & 0xFF) << 16;

		/* bit 24-31: ctrl byte
		** bit 24
		**
		**
		**			*/
		ctrl_byte = mod_no / 2;
#if !defined(SPI2STEP)
		ctrl_byte |= MOD_SPI_CTRL_START;
#endif
		ctrl_byte |= MOD_SPI_CTRL_CHAIN;	/* always chain */
		data |= ctrl_byte << 24;

	}else if (type == MOD_TYPE_FXS){

		/* bit 8-15: register byte */
		reg = reg & 0x7F;
		reg |= MOD_SPI_ADDR_FXS_WRITE; 
		data |= (reg & 0xFF) << 8;
		
		/* bit 16-23: chip select byte
		** bit 16
		**
		**
		**			*/
		if (mod_no % 2){
			/* Select first chip in a chain */
			chip_select = MOD_SPI_CS_FXS_CHIP_0;
		}else{
			/* Select second chip in a chain */
			chip_select = MOD_SPI_CS_FXS_CHIP_1;
		}
		data |= chip_select << 16;

		/* bit 24-31: ctrl byte
		** bit 24
		**
		**
		**			*/
		ctrl_byte = mod_no / 2;
#if !defined(SPI2STEP)
		ctrl_byte |= MOD_SPI_CTRL_START;
#endif
		ctrl_byte |= MOD_SPI_CTRL_FXS;
		if (chain){
			ctrl_byte |= MOD_SPI_CTRL_CHAIN;
		}
		data |= ctrl_byte << 24;

	}else{
		DEBUG_EVENT("%s: Module %d: Unsupported module type %d!\n",
				fe->name, mod_no, type);
		return -EINVAL;
	}

	card->hw_iface.bus_write_4(	card->hw,
					SPI_INTERFACE_REG,
					data);	
#if defined(SPI2STEP)
	WP_DELAY(1);
	data |= MOD_SPI_START;
	card->hw_iface.bus_write_4(	card->hw,
					SPI_INTERFACE_REG,
					data);	
#endif
#if 0
	DEBUG_EVENT("%s: %s: Module %d - Execute SPI command %08X\n",
					fe->name,
					__FUNCTION__,
					mod_no,
					data);
#endif

	for (i=0;i<10;i++){	
		WP_DELAY(10);
		card->hw_iface.bus_read_4(	card->hw,
						SPI_INTERFACE_REG,
						&data);

		if (data & MOD_SPI_BUSY){
			continue;
		}
	}

	if (data & MOD_SPI_BUSY) {
		DEBUG_EVENT("%s: Module %d: Critical Error (%s:%d)!\n",
					fe->name, mod_no,
					__FUNCTION__,__LINE__);
		return -EINVAL;
	}
	return 0;
}

//static void
void wp_remora_setreg(sdla_fe_t *fe, int mod_no, u8 reg, u8 value)
{
	sdla_t		*card = (sdla_t*)fe->card;
	//wan_smp_flag_t	flags, smp_flags;

	WAN_ASSERT1(card == NULL);

//	card->hw_iface.hw_lock(card->hw,&smp_flags);
//	wan_spin_lock_irq(&card->wandev.lock,&flags);

	__wp_remora_setreg(	fe,
				mod_no,
				fe->fe_param.remora.mod[mod_no].type,
				fe->fe_param.remora.mod[mod_no].chain,
				reg, value);

//	wan_spin_unlock_irq(&card->wandev.lock,&flags);
//	card->hw_iface.hw_unlock(card->hw,&smp_flags);
	return;
}

static u8
__wp_remora_getreg(sdla_fe_t *fe, int mod_no, int type, int chain, u8 reg)
{
	sdla_t		*card = (sdla_t*)fe->card;
	unsigned int	data = 0;
	unsigned char	chip_select = 0x00, ctrl_byte = 0x00;
	int		i;

	WAN_ASSERT(card == NULL);
	WAN_ASSERT(card->hw_iface.bus_write_4 == NULL);
	WAN_ASSERT(card->hw_iface.bus_read_4 == NULL);

	/* bit 0-7: data byte */
	data = 0x00;
	if (type == MOD_TYPE_FXO){

		/* bit 8-15: register byte */
		data |= (reg & 0xFF) << 8;

		/* bit 16-23: chip select byte
		** bit 16
		**
		**
		**			*/
		chip_select = 0x20;
		chip_select |= MOD_SPI_CS_FXO_READ;
		if (mod_no % 2 == 0){
			/* Select second chip in a chain */
			chip_select |= MOD_SPI_CS_FXO_CHIP_1;
		}
		data |= (chip_select & 0xFF) << 16;

		/* bit 24-31: ctrl byte
		** bit 24
		**
		**
		**			*/
		ctrl_byte = mod_no / 2;
#if !defined(SPI2STEP)
		ctrl_byte |= MOD_SPI_CTRL_START;
#endif
		ctrl_byte |= MOD_SPI_CTRL_CHAIN;	/* always chain */
		data |= ctrl_byte << 24;

	}else if (type == MOD_TYPE_FXS){

		/* bit 8-15: register byte */
		reg = reg & 0x7F;
		reg |= MOD_SPI_ADDR_FXS_READ; 
		data |= (reg & 0xFF) << 8;
		
		/* bit 16-23: chip select byte
		** bit 16
		**
		**
		**			*/
		if (mod_no % 2){
			/* Select first chip in a chain */
			chip_select = MOD_SPI_CS_FXS_CHIP_0;
		}else{
			/* Select second chip in a chain */
			chip_select = MOD_SPI_CS_FXS_CHIP_1;
		}
		data |= chip_select << 16;

		/* bit 24-31: ctrl byte
		** bit 24
		**
		**
		**			*/
		ctrl_byte = mod_no / 2;
#if !defined(SPI2STEP)
		ctrl_byte |= MOD_SPI_CTRL_START;
#endif
		ctrl_byte |= MOD_SPI_CTRL_FXS;
		if (chain){
			ctrl_byte |= MOD_SPI_CTRL_CHAIN;
		}
		data |= ctrl_byte << 24;

	}else{
		DEBUG_EVENT("%s: Module %d: Unsupported module type %d!\n",
				fe->name, mod_no, type);
		return -EINVAL;
	}

	card->hw_iface.bus_write_4(	card->hw,
					SPI_INTERFACE_REG,
					data);	
#if defined(SPI2STEP)
	WP_DELAY(1);
	data |= MOD_SPI_START;
	card->hw_iface.bus_write_4(	card->hw,
					SPI_INTERFACE_REG,
					data);	
#endif
#if 0
	DEBUG_EVENT("%s: %s: Module %d - Execute SPI command %08X\n",
					fe->name,
					__FUNCTION__,
					mod_no,
					data);
#endif
	for (i=0;i<10;i++){
		WP_DELAY(10);
		card->hw_iface.bus_read_4(	card->hw,
						SPI_INTERFACE_REG,
						&data);
		if (data & MOD_SPI_BUSY) {
			continue;
		}
	}

	if (data & MOD_SPI_BUSY){
		DEBUG_EVENT("%s: Module %d: Critical Error (%s:%d)!\n",
					fe->name, mod_no,
					__FUNCTION__,__LINE__);
		return 0xFF;
	}
	return (u8)(data & 0xFF);
}

//static u8
u8 wp_remora_getreg(sdla_fe_t *fe, int mod_no, u8 reg)
{
	sdla_t		*card = (sdla_t*)fe->card;
	//wan_smp_flag_t	flags, smp_flags;
	u8		value;

	WAN_ASSERT(card == NULL);

//	card->hw_iface.hw_lock(card->hw,&smp_flags);
//	wan_spin_lock_irq(&card->wandev.lock,&flags);
	
	value = __wp_remora_getreg(	fe,
					mod_no,
					fe->fe_param.remora.mod[mod_no].type,
					fe->fe_param.remora.mod[mod_no].chain,
					reg);

//	wan_spin_unlock_irq(&card->wandev.lock,&flags);
//	card->hw_iface.hw_unlock(card->hw,&smp_flags);
	return value;
}

static int
wp_proslic_setreg_indirect(sdla_fe_t *fe, int mod_no, unsigned char address, unsigned short data)
{

	wp_remora_setreg(fe, mod_no, IDA_LO,(unsigned char)(data & 0xFF));
	wp_remora_setreg(fe, mod_no, IDA_HI,(unsigned char)((data & 0xFF00)>>8));
	wp_remora_setreg(fe, mod_no, IAA,address);
	return 0;
}

static int
wp_proslic_getreg_indirect(sdla_fe_t *fe, int mod_no, unsigned char address)
{ 
	int res = -1;
	unsigned char data1, data2;

	wp_remora_setreg(fe, mod_no, IAA, address);
	data1 = wp_remora_getreg(fe, mod_no, IDA_LO);
	data2 = wp_remora_getreg(fe, mod_no, IDA_HI);
	res = data1 | (data2 << 8);
	return res;
}

static int wp_proslic_init_indirect_regs(sdla_fe_t *fe, int mod_no)
{
	unsigned char i;

	for (i=0; i<sizeof(indirect_regs) / sizeof(indirect_regs[0]); i++){
		if(wp_proslic_setreg_indirect(fe, mod_no, indirect_regs[i].address,indirect_regs[i].initial))
			return -1;
	}

	return 0;
}

static int wp_proslic_verify_indirect_regs(sdla_fe_t *fe, int mod_no)
{ 
	int passed = 1;
	unsigned short i, initial;
	int j;

	for (i=0; i<sizeof(indirect_regs) / sizeof(indirect_regs[0]); i++){ 
		
		j = wp_proslic_getreg_indirect(
				fe,
				mod_no,
				(unsigned char) indirect_regs[i].address);
		if (j < 0){
			DEBUG_EVENT("%s: Failed to read indirect register %d\n",
						fe->name,
						i);
			return -1;
		}
		initial= indirect_regs[i].initial;

		if ( j != initial && indirect_regs[i].altaddr != 255){
			DEBUG_EVENT(
			"%s: !!!!!!! %s  iREG %X = %X  should be %X\n",
						fe->name,
						indirect_regs[i].name,
						indirect_regs[i].address,
						j, initial);
			 passed = 0;
		}	
	}

	if (!passed) {
		DEBUG_EVENT(
		"%s: !!!!! Init Indirect Registers UNSUCCESSFULLY.\n",
						fe->name);
		return -1;
	}
	return 0;
}


static int wp_remora_chain_enable(sdla_fe_t *fe)
{
	int		mod_no;
	unsigned char	byte;

	for(mod_no = 0;mod_no < MAX_REMORA_MODULES; mod_no += 2){
		byte = __wp_remora_getreg(fe, mod_no, MOD_TYPE_FXS, 0, 0);
		if (byte != 0x05){
			DEBUG_RM("%s: Module %d/%d FXO\n",
						fe->name, mod_no,mod_no+1);
			fe->fe_param.remora.mod[mod_no].type	= MOD_TYPE_FXO;
			fe->fe_param.remora.mod[mod_no].chain	= MOD_CHAIN_ENABLED;
			fe->fe_param.remora.mod[mod_no+1].type	= MOD_TYPE_FXO;
			fe->fe_param.remora.mod[mod_no+1].chain	= MOD_CHAIN_ENABLED;
		}else{
			__wp_remora_setreg(fe, mod_no, MOD_TYPE_FXS, 0, 0, 0xC0);
			byte = __wp_remora_getreg(fe, mod_no, MOD_TYPE_FXS, 1, 0);
			if ((byte & 0x80) == 0x80){
				DEBUG_RM("%s: Module %d/%d FXS (chain)\n",
						fe->name, mod_no,mod_no+1);
				fe->fe_param.remora.mod[mod_no].type	= MOD_TYPE_FXS;
				fe->fe_param.remora.mod[mod_no].chain	= MOD_CHAIN_ENABLED;
				fe->fe_param.remora.mod[mod_no+1].type	= MOD_TYPE_FXS;
				fe->fe_param.remora.mod[mod_no+1].chain	= MOD_CHAIN_ENABLED;
			}else{
				DEBUG_EVENT(
				"%s:%d: Failed to enable chain mode (%02X)!\n", 
						fe->name, mod_no, byte);
				return -EINVAL;
			}
		}
	}

	/* Reset SPI interface */
	wp_remora_reset_spi(fe);

	/* Now enable chain mode for only FXS modules (FXO by default chain) */
	for(mod_no = 0;mod_no < MAX_REMORA_MODULES; mod_no += 2){
		if (fe->fe_param.remora.mod[mod_no].type == MOD_TYPE_FXS){
			__wp_remora_setreg(fe,mod_no,MOD_TYPE_FXS,0,0,0xC0);
		}else{
			byte = __wp_remora_getreg(fe,mod_no,MOD_TYPE_FXO,1,2);
			if (byte == 0x03){
				fe->fe_param.remora.mod[mod_no].type	= MOD_TYPE_FXO;
				fe->fe_param.remora.mod[mod_no+1].type	= MOD_TYPE_FXO;
			}else{
				fe->fe_param.remora.mod[mod_no].type	= MOD_TYPE_NONE;
				fe->fe_param.remora.mod[mod_no+1].type	= MOD_TYPE_NONE;
			}
		}
	}
	return 0;
}

static int wp_init_proslic_insane(sdla_fe_t *fe, int mod_no)
{
	unsigned char	value;

	value = wp_remora_getreg(fe, mod_no, 0);
	if ((value & 0x30) >> 4){
		DEBUG_TEST("%s: Proslic on module %d is not a Si3210!\n",
					fe->name,
					mod_no);
		return -1;
	}
	if (((value & 0x0F) == 0) || ((value & 0x0F) == 0x0F)){
		DEBUG_TEST("%s: Proslic is not loaded!\n",
					fe->name);
		return -1;
	}
	if ((value & 0x0F) < 2){
		DEBUG_EVENT("%s: Proslic 3210 version %d is too old!\n",
					fe->name,
					value & 0x0F);
		return -1;
	}

	value = wp_remora_getreg(fe, mod_no, 8);
	if (value != 0x2) {
		DEBUG_EVENT(
		"%s: Proslic on module %d insane (1) %d should be 2!\n",
					fe->name, mod_no, value);
		return -1;
	}

	value = wp_remora_getreg(fe, mod_no, 64);
	if (value != 0x0) {
		DEBUG_EVENT(
		"%s: Proslic on modyle %d insane (2) %d should be 0!\n",
					fe->name, mod_no, value);
		return -1;
	} 

	value = wp_remora_getreg(fe, mod_no, 11);
	if (value != 0x33) {
		DEBUG_EVENT(
		"%s: Proslic on module %d insane (3) %02X should be 0x33\n",
					fe->name, mod_no, value);
		return -1;
	}
	wp_remora_setreg(fe, mod_no, 30, 0);
	return 0;
}


static int wp_powerup_proslic(sdla_fe_t *fe, int mod_no)
{
	unsigned long	ticks;
	int		loopcurrent = 20, lim;
	unsigned char	vbat;
	
	/* powerup */ 
	wp_remora_setreg(fe, mod_no, 93, 0x1F);
	/* set the period of the DC-DC converter to 1/64 kHz  START OUT SLOW*/
	wp_remora_setreg(fe, mod_no, 92, 0xf5);

	ticks = SYSTEM_TICKS;
	wp_remora_setreg(fe, mod_no, 14, 0x0);	/* DIFF DEMO 0x10 */
	while((vbat = wp_remora_getreg(fe, mod_no, 82)) < 0xC0){
		/* Wait no more than 500ms */
		if ((SYSTEM_TICKS - ticks) > HZ/2){
			break;
		}
	}
//ALEXALEX
	if (vbat < 0xc0){
		if (fe->fe_param.remora.mod[mod_no].u.fxs.proslic_power == PROSLIC_POWER_UNKNOWN){
			DEBUG_EVENT(
			"%s: Module %d: Failed to powerup within %d ms (%d mV only)!\n",
					fe->name,
					mod_no,
					(int)(((SYSTEM_TICKS - ticks) * 1000 / HZ)),
					vbat * 375);
			DEBUG_EVENT(
			"%s: Module %d: Did you remember to plug in the power cable?\n",
					fe->name,
					mod_no);

		}
		fe->fe_param.remora.mod[mod_no].u.fxs.proslic_power = PROSLIC_POWER_WARNED;
		return -1;
	}
	fe->fe_param.remora.mod[mod_no].u.fxs.proslic_power = PROSLIC_POWER_ON;

// NEW VVV
        /* Proslic max allowed loop current, reg 71 LOOP_I_LIMIT */
        /* If out of range, just set it to the default value     */
        lim = (loopcurrent - 20) / 3;
        if ( loopcurrent > 41 ) {
                lim = 0;
		DEBUG_RM(
		"%s: Module %d: Loop current out of range (default 20mA)!\n",
					fe->name, mod_no);
        }else{
		DEBUG_RM("%s: Loop current set to %dmA!\n",
					fe->name,
					(lim*3)+20);
	}
        wp_remora_setreg(fe, mod_no, 71,lim);
// NEW ^^^ 

	wp_remora_setreg(fe, mod_no, 93, 0x99);  /* DC-DC Calibration  */
#if 1
	// Wait for DC-DC Calibration to complete
	ticks = SYSTEM_TICKS;
	while(0x80 & wp_remora_getreg(fe, mod_no, 93)){
		if ((SYSTEM_TICKS - ticks) > 2*HZ){
			DEBUG_EVENT(
			"%s: Module %d: Timeout waiting for DC-DC calibration\n",
						fe->name,
						mod_no);
			return -EINVAL;
		}
	}
#endif
	return 0;
}

static int wp_proslic_powerleak_test(sdla_fe_t *fe, int mod_no)
{
	unsigned long	ticks;
	unsigned char	vbat;

	/* powerleak */ 
	wp_remora_setreg(fe, mod_no, 64, 0);
	wp_remora_setreg(fe, mod_no, 14, 0x10);
	/* wait for 1 s */
	ticks = SYSTEM_TICKS;
	while((vbat = wp_remora_getreg(fe, mod_no, 82)) >= 0x6){
		if ((SYSTEM_TICKS - ticks) > (HZ/2)){
			break;
		}
	}
	if (vbat < 0x6){
		DEBUG_EVENT(
		"%s: Module %d: Excessive leakage detected: %d volts (%02x) after %d ms\n",
					fe->name, mod_no,
					376 * vbat / 1000,
					vbat,
					(int)((SYSTEM_TICKS - ticks) * 1000 / HZ));
		return -1;
	}
	DEBUG_RM("%s: Module %d: Post-leakage voltage: %d volts\n",
					fe->name,
					mod_no,
					376 * vbat / 1000);

	return 0;
}


//static
int wp_init_proslic(sdla_fe_t *fe, int mod_no, int fast, int sane)
{
	unsigned long	ticks;
	unsigned short	tmp[5];
	unsigned char	value, int_enable3 = 0x00, int_enable2 = 0x00;
	int		i, x;

	/* By default, don't send on hook */
	if (fe->fe_cfg.cfg.remora.reversepolarity){
		fe->fe_param.remora.mod[mod_no].u.fxs.idletxhookstate = 5;
	}else{
		fe->fe_param.remora.mod[mod_no].u.fxs.idletxhookstate = 1;
	}

	/* Step 8 */
	if (!sane && wp_init_proslic_insane(fe, mod_no)){
		return -2;
	}

	if (sane){
		wp_remora_setreg(fe, mod_no, 14, 0x10);
	}

	/* Step 9 */
	if (wp_proslic_init_indirect_regs(fe, mod_no)) {
		DEBUG_EVENT(
		"%s: Module %d: Indirect Registers failed to initialize!\n",
							fe->name,
							mod_no);
		return -1;
	}
	wp_proslic_setreg_indirect(fe, mod_no, 97,0);

	/* Steo 10 */
	wp_remora_setreg(fe, mod_no, 8, 0);		//DIGIUM
	wp_remora_setreg(fe, mod_no, 108, 0xeb);	// DIGIUM
	wp_remora_setreg(fe, mod_no, 67, 0x17);
	wp_remora_setreg(fe, mod_no, 66, 1);

	/* Flush ProSLIC digital filters by setting to clear, while
	** saving old values */
	for (x=0;x<5;x++) {
		tmp[x] = wp_proslic_getreg_indirect(fe, mod_no, x + 35);
		wp_proslic_setreg_indirect(fe, mod_no, x + 35, 0x8000);
	}

	/* Power up the DC-DC converter */
	if (wp_powerup_proslic(fe, mod_no)) {
		DEBUG_EVENT(
		"%s: Module %d: Unable to do INITIAL ProSLIC powerup!\n",
					fe->name,
					mod_no);
		return -1;
	}

	if (!fast){
		if (wp_proslic_powerleak_test(fe, mod_no)){
			DEBUG_EVENT(
			"%s: Module %d: Proslic failed leakge the short circuit\n",
						fe->name,
						mod_no);
		}

		/* Step 12 */
		if (wp_powerup_proslic(fe, mod_no)) {
			DEBUG_EVENT(
			"%s: Module %d: Unable to do INITIAL ProSLIC powerup!\n",
						fe->name,
						mod_no);
			return -1;
		}

		/* Step 13 */
		wp_remora_setreg(fe, mod_no, 64, 0);

		/* Step 14 */
		wp_remora_setreg(fe, mod_no, 97, 0x1E);
		wp_remora_setreg(fe, mod_no, 96, 0x47);

		/* Step 15 */
		ticks = SYSTEM_TICKS;
		while(wp_remora_getreg(fe, mod_no, 96) != 0){
			if ((SYSTEM_TICKS - ticks) > 400){
				DEBUG_EVENT(
				"%s: Module %d: Timeout on SLIC calibration (15)!\n",
						fe->name, mod_no);
				return -1;
			}
		}

		/* Step 16 */
		/* Insert manual calibration for sangoma Si3210 */
		wp_remora_setreg(fe, mod_no, 98, 0x10);
		for (i = 0x1f; i > 0; i--){

			wp_remora_setreg(fe, mod_no, 98, i);
			wait_just_a_bit(4);
			//ticks = SYSTEM_TICKS;
			//while((SYSTEM_TICKS - ticks) < 4);
			if ((wp_remora_getreg(fe, mod_no, 88)) == 0){
				break;
			}
		} // for

		wp_remora_setreg(fe, mod_no, 99, 0x10);
		for (i = 0x1f; i > 0; i--){

			wp_remora_setreg(fe, mod_no, 99, i);
			wait_just_a_bit(4);
			//ticks = SYSTEM_TICKS;
			//while((SYSTEM_TICKS - ticks) < 4);
			if ((wp_remora_getreg(fe, mod_no, 89)) == 0){
				break;
			}
		}//for

		/* Step 17 */
		value = wp_remora_getreg(fe, mod_no, 23);
		wp_remora_setreg(fe, mod_no, 23, value | 0x04);

		/* Step 18 */
		// DAC offset and without common mode calibration.
		wp_remora_setreg(fe, mod_no, 97, 0x01/*0x18*/);	/* Manual after */
		// Calibrate common mode and differential DAC mode DAC + ILIM
		wp_remora_setreg(fe, mod_no, 96, 0x40/*0x47*/);

		/* Step 19 */
		ticks = SYSTEM_TICKS;
		while(wp_remora_getreg(fe, mod_no, 96) != 0){
			if ((SYSTEM_TICKS - ticks) > 2000/*400*/){
				DEBUG_EVENT(
				"%s: Module %d: Timeout on SLIC calibration (19:%02X)!\n",
						fe->name,
						mod_no,
						wp_remora_getreg(fe, mod_no, 96));
				return -1;
			}
		}
		DEBUG_RM("%s: SLIC calibration complete (%ld)\n",
					fe->name, SYSTEM_TICKS-ticks);
		/* Save calibration vectors */
		for (x=0;x<NUM_CAL_REGS;x++){
			fe->fe_param.remora.mod[mod_no].u.fxs.callregs.vals[x] =
					wp_remora_getreg(fe, mod_no, 96 + x);
		}

	}else{

		/* Save calibration vectors */
		for (x=0;x<NUM_CAL_REGS;x++){
			wp_remora_setreg(fe, mod_no, 96 + x, 
				fe->fe_param.remora.mod[mod_no].u.fxs.callregs.vals[x]);
		}

	}    

	/* Step 20 */
	wp_proslic_setreg_indirect(fe, mod_no, 88, 0);
	wp_proslic_setreg_indirect(fe, mod_no, 89, 0);
	wp_proslic_setreg_indirect(fe, mod_no, 90, 0);
	wp_proslic_setreg_indirect(fe, mod_no, 91, 0);
	wp_proslic_setreg_indirect(fe, mod_no, 92, 0);
	wp_proslic_setreg_indirect(fe, mod_no, 93, 0);
	wp_proslic_setreg_indirect(fe, mod_no, 94, 0);
	wp_proslic_setreg_indirect(fe, mod_no, 95, 0);

	wp_remora_setreg(fe, mod_no, 21, 0);//(0) Disable all interupts in DR21
	wp_remora_setreg(fe, mod_no, 22, 0);//(0) Disable all interupts in DR21
	wp_remora_setreg(fe, mod_no, 23, 0);//(0) Disabel all interupts in DR21
	wp_remora_setreg(fe, mod_no, 64, 0);//(0)

	/* IM HERE*/

	wp_remora_setreg(fe, mod_no, 64, 0x1);

	value = wp_remora_getreg(fe, mod_no, 68); 
	//FIXME ???value = value & 0x03;
	//if (value & 4){
	//	printf("Module %d Timeout!\n", mod_no);
	//	return -1;
	//}

#if 1
	wp_remora_setreg(fe, mod_no, 64, 0x00);

	// this is a singular calibration bit for longitudinal calibration
	wp_remora_setreg(fe, mod_no, 97, 0x01);
	wp_remora_setreg(fe, mod_no, 96, 0x40);

	value = wp_remora_getreg(fe, mod_no, 96); 

	wp_remora_setreg(fe, mod_no, 18,0xff);
	wp_remora_setreg(fe, mod_no, 19,0xff);
	wp_remora_setreg(fe, mod_no, 20,0xff);

	//wp_remora_setreg(fe, mod_no, 64,0x1);
#endif

	/* Perform DC-DC calibration */
	wp_remora_setreg(fe, mod_no,  93, 0x99);
	wait_just_a_bit(10);
	value = wp_remora_getreg(fe, mod_no, 107);
	if ((value < 0x2) || (value > 0xd)) {
		DEBUG_EVENT(
		"%s: Module %d: DC-DC calibration has a surprising direct 107 of 0x%02x!\n",
					fe->name,
					mod_no,
					value);
		wp_remora_setreg(fe, mod_no,  107, 0x8);
	}

	for (x=0;x<5;x++) {
		wp_proslic_setreg_indirect(fe, mod_no, x + 35, tmp[x]);
	}

	if (wp_proslic_verify_indirect_regs(fe, mod_no)) {
		DEBUG_EVENT(
		"%s: Module %d: Indirect Registers failed verification.\n",
					fe->name,
					mod_no);
		return -1;
	}

	wp_remora_setreg(fe, mod_no, 1, 0x28);
	/* U-Law 8-bit interface */
	/* Tx Start count low byte  0 */
	wp_remora_setreg(fe, mod_no, 2, mod_no * 8 + 1);
	/* Tx Start count high byte 0 */
	wp_remora_setreg(fe, mod_no, 3, 0);
	/* Rx Start count low byte  0 */
	wp_remora_setreg(fe, mod_no, 4, mod_no * 8 + 1);
	/* Rx Start count high byte 0 */
	wp_remora_setreg(fe, mod_no, 5, 0);
	/* Clear all interrupt */
	wp_remora_setreg(fe, mod_no, 18, 0xff);
	wp_remora_setreg(fe, mod_no, 19, 0xff);
	wp_remora_setreg(fe, mod_no, 20, 0xff);
	wp_remora_setreg(fe, mod_no, 73, 0x04);

	
	//wp_remora_setreg(fe, mod_no, 21, 0x00); // enable interrupt
	int_enable2 = 0x00;
#if defined(WAN_REMORA_FXS_LCIP)
	/* Loop detection interrupt */
	DEBUG_EVENT("%s: Module %d: Enable Loop detection interrupt!\n",
				fe->name, mod_no);
	int_enable2 |= 0x02;
#endif
	wp_remora_setreg(fe, mod_no, 22, int_enable2);

	int_enable3 = 0x00;
#if defined(WAN_REMORA_FXS_DTMF)
	int_enable3 |= 0x01;		
	/* DTMF detection interrupt */
	DEBUG_EVENT("%s: Module %d: Enable DTMF detection interrupt!\n",
				fe->name, mod_no);
#endif
	wp_remora_setreg(fe, mod_no, 23, int_enable3);

#if 0
	/* Enable loopback */
	wp_remora_setreg(fe, mod_no, 8,  0x2);
	wp_remora_setreg(fe, mod_no, 14, 0x0);
	wp_remora_setreg(fe, mod_no, 64, 0x0);
	wp_remora_setreg(fe, mod_no, 1,  0x08);
#endif

	wp_remora_setreg(fe, mod_no, 64, 0x1);

	return 0;
}

static int wp_voicedaa_insane(sdla_fe_t *fe, int mod_no)
{
	unsigned char	byte;

	byte = wp_remora_getreg(fe, mod_no, 2);
	if (byte != 0x3)
		return -2;
	byte = wp_remora_getreg(fe, mod_no, 11);
	DEBUG_TEST("%s: Module %d: VoiceDAA System: %02x\n",
				fe->name,
				mod_no,
				byte & 0xf);
	return 0;
}


static int wp_init_voicedaa(sdla_fe_t *fe, int mod_no, int sane)
{
	unsigned char reg16=0, reg26=0, reg30=0, reg31=0;
	unsigned long	ticks;

	if (!sane && wp_voicedaa_insane(fe, mod_no)){
		return -2;
	}

	/* Software reset */
	wp_remora_setreg(fe, mod_no, 1, 0x80);

	/* Wait just a bit */
	wait_just_a_bit(HZ/10);

	/* Enable PCM, ulaw */
	switch(fe->fe_cfg.tdmv_law){
	case WAN_TDMV_ALAW:
		wp_remora_setreg(fe, mod_no, 33, 0x20);
		break;
	case WAN_TDMV_MULAW:
		wp_remora_setreg(fe, mod_no, 33, 0x28);
		break;
	}

	/* Set On-hook speed, Ringer impedence, and ringer threshold */
	reg16 |= (fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].ohs << 6);
	reg16 |= (fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].rz << 1);
	reg16 |= (fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].rt);
	wp_remora_setreg(fe, mod_no, 16, reg16);
	
	/* Set DC Termination:
	**	Tip/Ring voltage adjust,
	**	minimum operational current,
	**	current limitation */
	reg26 |= (fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].dcv << 6);
	reg26 |= (fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].mini << 4);
	reg26 |= (fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].ilim << 1);
	wp_remora_setreg(fe, mod_no, 26, reg26);

	/* Set AC Impedence */
	reg30 = (fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].acim);
	wp_remora_setreg(fe, mod_no, 30, reg30);

	/* Misc. DAA parameters */
	reg31 = 0xa3;
	reg31 |= (fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].ohs2 << 3);
	wp_remora_setreg(fe, mod_no, 31, reg31);

	/* Set Transmit/Receive timeslot */
	wp_remora_setreg(fe, mod_no, 34, mod_no * 8 + 1);
	wp_remora_setreg(fe, mod_no, 35, 0x00);
	wp_remora_setreg(fe, mod_no, 36, mod_no * 8 + 1);
	wp_remora_setreg(fe, mod_no, 37, 0x00);

	/* Enable ISO-Cap */
	wp_remora_setreg(fe, mod_no, 6, 0x00);

	/* Wait 1000ms for ISO-cap to come up */
	ticks = SYSTEM_TICKS + 2*HZ;
	while(!(wp_remora_getreg(fe, mod_no, 11) & 0xf0)){
		if (SYSTEM_TICKS > ticks){
			break;
		}
		wait_just_a_bit(HZ/10);
	}

	if (!(wp_remora_getreg(fe, mod_no, 11) & 0xf0)) {
		DEBUG_EVENT(
		"%s: Module %d: VoiceDAA did not bring up ISO link properly!\n",
					fe->name,
					mod_no);
		return -1;
	}
	DEBUG_TEST("%s: Module %d: ISO-Cap is now up, line side: %02x rev %02x\n", 
			fe->name,
			mod_no,
			wp_remora_getreg(fe, mod_no, 11) >> 4,
			(wp_remora_getreg(fe, mod_no, 13) >> 2) & 0xf);

	/* Enable on-hook line monitor */
	wp_remora_setreg(fe, mod_no, 5, 0x08);

	/* NZ -- crank the tx gain up by 7 dB */
	if (!strcmp(fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].name, "NEWZEALAND")) {
		DEBUG_EVENT("%s: Module %d: Adjusting gain\n",
					fe->name,
					mod_no);
		wp_remora_setreg(fe, mod_no, 38, 0x7);
	}

	return 0;
}

/******************************************************************************
** wp_remora_config() - 
**
**	OK
*/
int wp_remora_config(void *pfe, void *pfe_iface)
{
	sdla_fe_t	*fe = (sdla_fe_t*)pfe;
	sdla_fe_iface_t	*fe_iface = (sdla_fe_iface_t*)pfe_iface;
	int		err, mod_no;

	DEBUG_EVENT("%s: Configuring FXS/FXO Front End ...\n",
        		     	fe->name);

	/* Reset SPI interface */
	wp_remora_reset_spi(fe);

	/* Search for installed modules and enable chain for all modules */
	if (wp_remora_chain_enable(fe)){
		DEBUG_EVENT("%s: Failed enable chain mode for all modules!\n",
				fe->name);
		return -EINVAL;
	}

	/* Auto detect FXS and FXO modules */
	for(mod_no = 0; mod_no < MAX_REMORA_MODULES; mod_no++){
		switch(fe->fe_param.remora.mod[mod_no].type){
		case MOD_TYPE_FXS:
			if (!(err = wp_init_proslic(fe, mod_no, 0, 0))){
				DEBUG_EVENT(
				"%s: Module %d: Installed -- Auto FXS!\n",
						fe->name,
						mod_no);	
				wan_set_bit(mod_no, &fe->fe_param.remora.module_map);
			}else{
				DEBUG_EVENT(
				"%s: Module %d: FXS failed!\n",
						fe->name,
						mod_no);	
			}
			break;

		case MOD_TYPE_FXO:
			err = wp_init_voicedaa(fe, mod_no, 0);
			if (!err){
				DEBUG_EVENT(
				"%s: Module %d: Installed -- Auto FXO (%s mode)!\n",
						fe->name,
						mod_no,
						fxo_modes[fe->fe_cfg.cfg.remora.fxo_opermode].name);
				wan_set_bit(mod_no, &fe->fe_param.remora.module_map);
			}else{
				DEBUG_EVENT(
				"%s: Module %d: FXO failed!\n",
						fe->name,
						mod_no);	
			}
			break;

		default:
			DEBUG_TDMV("%s: Module %d: Not Installed!\n",
						fe->name,
						mod_no);	
			break;
		}
	}
	fe_iface->isr			= &wp_remora_intr;
	fe_iface->check_isr		= &wp_remora_check_intr;

	return 0;
}

int wp_remora_unconfig(void *pfe)
{
	sdla_fe_t	*fe = (sdla_fe_t*)pfe;

	DEBUG_EVENT("%s: Unconfiguring FXS/FXO Front End...\n",
        		     	fe->name);
	return 0;
}

static int wp_remora_check_intr_fxs(sdla_fe_t *fe, int mod_no)
{
	unsigned char	stat1 = 0x0, stat2 = 0x00, stat3 = 0x00;

	if (wp_remora_getreg(fe, mod_no, 8)){
		DEBUG_EVENT("Ouch, part reset, quickly restoring reality (%d)\n",
				fe->name, mod_no);
		wp_init_proslic(fe, mod_no, 1, 0);
		return 0;
	}

	stat1 = wp_remora_getreg(fe, mod_no, 18);
	stat2 = wp_remora_getreg(fe, mod_no, 19);
	stat3 = wp_remora_getreg(fe, mod_no, 20);

	if (stat1 || stat2 || stat3){
		DEBUG_EVENT("%s: Module %d: Interrupt status %02X %02X %02X\n",
				fe->name, mod_no, stat1, stat2, stat3);
		return 1;
	}
	if (stat1 & 0x2){
		return 1;
	}
	if (stat2 & 0x1){
		return 1;
	}
	return 0;	
}

static int wp_remora_check_intr_fxo(sdla_fe_t *fe, int mod_no)
{
	return 0;	
}

static int wp_remora_check_intr(sdla_fe_t *fe)
{
	int		mod_no = 0, pending = 0;

	for(mod_no = 0; mod_no < MAX_REMORA_MODULES; mod_no++){
		if (!wan_test_bit(mod_no, &fe->fe_param.remora.module_map)) {
			continue;
		}
		if (fe->fe_param.remora.mod[mod_no].type == MOD_TYPE_FXS){
			pending = wp_remora_check_intr_fxs(fe, mod_no);
		}else if (fe->fe_param.remora.mod[mod_no].type == MOD_TYPE_FXO){
			pending = wp_remora_check_intr_fxo(fe, mod_no);
		}
		if (pending) return pending;
		
	}
	return 0;
} 

static int wp_remora_intr_fxs(sdla_fe_t *fe, int mod_no)
{
	sdla_t		*card = NULL;
	unsigned char	stat1 = 0x0, stat2 = 0x00, stat3 = 0x00;


	WAN_ASSERT(fe->card == NULL);
	card		= fe->card;

	stat1 = wp_remora_getreg(fe, mod_no, 18);
	stat2 = wp_remora_getreg(fe, mod_no, 19);
	stat3 = wp_remora_getreg(fe, mod_no, 20);

	if (stat1){
		/* Ack interrupts for now */
		wp_remora_setreg(fe, mod_no, 18, stat1);
	}
	if (stat2){
		if (stat2 & 0x02){
			unsigned char	status = 0x00;
			status = wp_remora_getreg(fe, mod_no, 68);
			DEBUG_EVENT(
			"%s: Module %d: LCIP interrupt pending!\n",
				fe->name, mod_no);
			if (status & 0x01){
				DEBUG_EVENT(
				"%s: Module %d: Off-hook status!\n",
						fe->name, mod_no);
				if (card->wandev.fe_notify_iface.hook_state){
					card->wandev.fe_notify_iface.hook_state(fe, mod_no, 1);
				}
				fe->fe_param.remora.mod[mod_no].u.fxs.oldrxhook = 1;
			}else if (fe->fe_param.remora.mod[mod_no].u.fxs.oldrxhook){
				DEBUG_EVENT(
				"%s: Module %d: On-hook status!\n",
						fe->name, mod_no);
				if (card->wandev.fe_notify_iface.hook_state){
					card->wandev.fe_notify_iface.hook_state(fe, mod_no, 0);
				}
				fe->fe_param.remora.mod[mod_no].u.fxs.oldrxhook = 0;
			}
		}
		if (stat2 & 0x01){
			DEBUG_EVENT(
			"%s: Module %d: Ring TRIP interrupt pending!\n",
				fe->name, mod_no);
			DEBUG_EVENT(
			"%s: Module %d: Reg[64]=%02X Reg[68]=%02X\n",
				fe->name, mod_no,
				wp_remora_getreg(fe,mod_no,64),
				wp_remora_getreg(fe,mod_no,68));
		}
		wp_remora_setreg(fe, mod_no, 19, stat2);
	}

	if (stat3){
		if (stat3 & 0x1){
			unsigned char	status;
			status = wp_remora_getreg(fe, mod_no, 24);
			if (status & 0x10){
				unsigned char	*digit = "Wrong";
				status &= 0xF;
				switch(status){
				case 0x01: digit = "1"; break;
				case 0x02: digit = "2"; break;
				case 0x03: digit = "3"; break;
				case 0x04: digit = "4"; break;
				case 0x05: digit = "5"; break;
				case 0x06: digit = "6"; break;
				case 0x07: digit = "7"; break;
				case 0x08: digit = "8"; break;
				case 0x09: digit = "9"; break;
				case 0x0A: digit = "0"; break;
				case 0x0B: digit = "*"; break;
				case 0x0C: digit = "#"; break;
				case 0x0D: digit = "A"; break;
				case 0x0E: digit = "B"; break;
				case 0x0F: digit = "C"; break;
				case 0x00: digit = "D"; break;
				}
				DEBUG_EVENT("%s: Module %d: TDMV digit %s!\n",
						fe->name,
						mod_no + 1,
						digit);
				wp_remora_setreg(fe, mod_no, 20, 0x1);
			}
		}
		wp_remora_setreg(fe, mod_no, 20, stat3);
	}
	return 0;
}

static int wp_remora_intr_fxo(sdla_fe_t *fe, int mod_no)
{
	return 0;
}

static int wp_remora_intr(sdla_fe_t *fe)
{
	int		mod_no = 0;

	for(mod_no = 0; mod_no < MAX_REMORA_MODULES; mod_no++){
		if (!wan_test_bit(mod_no, &fe->fe_param.remora.module_map)) {
			continue;
		}
		if (fe->fe_param.remora.mod[mod_no].type == MOD_TYPE_FXS){
			wp_remora_intr_fxs(fe, mod_no);
		}else if (fe->fe_param.remora.mod[mod_no].type == MOD_TYPE_FXO){
			wp_remora_intr_fxo(fe, mod_no);
		}
	}
	return 0;
}
