#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#if defined(__LINUX__)
# include <linux/if.h>
# include <linux/if_packet.h>
# include <linux/wanpipe_defines.h>
# include <linux/sdlasfm.h>
# include <linux/wanpipe_cfg.h>
#else
# include <net/if.h>
# include <net/wanpipe_defines.h>
# include <net/sdlasfm.h>
# include <net/wanpipe_cfg.h>
#endif

#include "wan_aft_prg.h"

#define JP8_VALUE               0x02
#define JP7_VALUE               0x01
#define SW0_VALUE               0x04
#define SW1_VALUE               0x08   

#define MASK_DEF_SECTOR_FLASH	0x00
#define MASK_USER_SECTOR_FLASH	0x08

#define WRITE_DEF_SECTOR_DSBL   0x01
#define FRONT_END_TYPE_MASK     0x38

#define AFT_BIT_DEV_ADDR_CLEAR	0x600
#define AFT_BIT_DEV_ADDR_CPLD	0x200

#define AFT4_BIT_DEV_ADDR_CLEAR	0x800 /* QUADR */
#define AFT4_BIT_DEV_ADDR_CPLD	0x800 /* QUADR */

#define BIT_A18_SECTOR_SA4_SA7	0x20

#define AFT_SHARK_USER_SECTOR_START_ADDR	0x80000

#define AFT_SHARK_CORE_SIZE	402936

/* Manufacturer code */
#define MCODE_ST	0x20

/* Device code */
#define DCODE_M29W040B	0xE3
#define DCODE_M29W800DT	0xD7
#define DCODE_M29W800DB	0x5B

#define M29W800DT_FID	0
#define M29W800DB_FID	1

struct {
	unsigned long	start_addr;
	unsigned long	len;
	unsigned int	sector_type;
} aft_shark_flash_spec [2][19] = 
	{
		{ 
			{ 0x00000, 0xFFFF, DEF_SECTOR_FLASH }, 
			{ 0x10000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x20000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x30000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x40000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x50000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x60000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x70000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x80000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0x90000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xA0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xB0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xC0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xD0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xE0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xF0000, 0x7FFF, USER_SECTOR_FLASH },
			{ 0xF8000, 0x1FFF, USER_SECTOR_FLASH },
			{ 0xFA000, 0x1FFF, USER_SECTOR_FLASH },
			{ 0xFC000, 0x3FFF, USER_SECTOR_FLASH },
		},

		{
			{ 0x00000, 0x3FFF, DEF_SECTOR_FLASH },
			{ 0x04000, 0x1FFF, DEF_SECTOR_FLASH },
			{ 0x06000, 0x1FFF, DEF_SECTOR_FLASH },
			{ 0x08000, 0x7FFF, DEF_SECTOR_FLASH },
			{ 0x10000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x20000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x30000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x40000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x50000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x60000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x70000, 0xFFFF, DEF_SECTOR_FLASH },
			{ 0x80000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0x90000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xA0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xB0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xC0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xD0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xE0000, 0xFFFF, USER_SECTOR_FLASH },
			{ 0xF0000, 0xFFFF, USER_SECTOR_FLASH },
		}
	};

aftup_flash_t	aft_shark_flash = {  0x014F, AFT_SHARK_CORE_SIZE };

extern int	verbose;

extern int	progress_bar(char*);
/*
 ******************************************************************************
			  FUNCTION PROTOTYPES
 ******************************************************************************
*/
		
extern int exec_read_cmd(void*,unsigned int, unsigned int, unsigned int*);
extern int exec_write_cmd(void*,unsigned int, unsigned int, unsigned int);
extern void hit_any_key(void);

static int aft_shark_reset_flash(wan_aft_cpld_t *cpld);
static int aft_shark_is_protected(wan_aft_cpld_t *cpld, int stype);
static int aft_shark_flash_id(wan_aft_cpld_t *cpld, int mtype, int stype, int *flash_id);
static int aft_shark_reload_flash(wan_aft_cpld_t *cpld, int sector_type);
static int aft_shark_prg_flash_byte(wan_aft_cpld_t*,int,unsigned long,unsigned char);
static unsigned char aft_shark_read_flash_byte(wan_aft_cpld_t*,int,int,unsigned long);
static int aft_shark_erase_flash(wan_aft_cpld_t*,int,int);

aftup_flash_iface_t aftup_shark_flash_iface = 
{
	aft_shark_reset_flash,
	aft_shark_is_protected,
	aft_shark_flash_id,
	aft_shark_reload_flash,
	aft_shark_prg_flash_byte,
	aft_shark_read_flash_byte,
	aft_shark_erase_flash
};

/******************************************************************************
*			  FUNCTION DEFINITION	
******************************************************************************/
static unsigned int
write_cpld(wan_aft_cpld_t *cpld, unsigned short cpld_off,unsigned short cpld_data)
{
	cpld_off |= AFT4_BIT_DEV_ADDR_CPLD; 
	exec_write_cmd(cpld->private, 0x46, 2, cpld_off);
	exec_write_cmd(cpld->private, 0x44, 2, cpld_data);
	return 0;
}

static unsigned int
read_cpld(wan_aft_cpld_t *cpld, unsigned short cpld_off)
{
	unsigned int cpld_data;
		
	cpld_off |= AFT4_BIT_DEV_ADDR_CPLD; 
	exec_write_cmd(cpld->private, 0x46, 2, cpld_off);
	if (exec_read_cmd(cpld->private, 0x44, 4, &cpld_data) == 0){
		return cpld_data;
	}else{
		return 0;
	}
}

static unsigned int
write_flash(wan_aft_cpld_t *cpld, int stype, int mtype, unsigned long off,unsigned char data)
{
	unsigned char	offset;

	//Writing flash address to cpld
	offset = off & 0xFF;
	write_cpld(cpld, 0x05, offset);
	offset = (off >> 8) & 0xFF;
	write_cpld(cpld, 0x06, offset);
	offset = (off >> 16) & 0x7;
	if (mtype == MEMORY_TYPE_SRAM){
		offset |= MASK_MEMORY_TYPE_SRAM;
	}else if (mtype == MEMORY_TYPE_FLASH){
		offset |= MASK_MEMORY_TYPE_FLASH;
	}else{
		return -EINVAL;
	}
	if (stype == USER_SECTOR_FLASH){
		offset |= MASK_USER_SECTOR_FLASH;
	}else if (stype == DEF_SECTOR_FLASH){
		;
	}else{
		return -EINVAL;
	}
	write_cpld(cpld, 0x07, offset);
	write_cpld(cpld, 0x04, data);
        write_cpld(cpld, 0x07, 0x00);  // disable CS signal for the Boot FLASH/SRAM

	return 0;
}

static unsigned char
aft_shark_read_flash_byte(wan_aft_cpld_t *cpld, int stype, int mtype, unsigned long off)
{
	unsigned char offset;
        unsigned char data;

	//Writing flash address to cpld
	offset = off & 0xFF;
	write_cpld(cpld, 0x05, offset);
	offset = (off >> 8) & 0xFF;
	write_cpld(cpld, 0x06, offset);
	offset = (off >> 16) & 0x7;
        offset |= MASK_MEMORY_TYPE_FLASH;
//           
//	if (memory_type == MEMORY_TYPE_SRAM){
//		offset |= MASK_MEMORY_TYPE_SRAM;
//	}else if (memory_type == MEMORY_TYPE_FLASH){
//		offset |= MASK_MEMORY_TYPE_FLASH;
//	}else{
//		return -EINVAL;
//	}
	if (stype == USER_SECTOR_FLASH){
		offset |= MASK_USER_SECTOR_FLASH;
	}else if (stype == DEF_SECTOR_FLASH){
		;
	}else{
		return -EINVAL;
	}

	write_cpld(cpld, 0x07, offset);
        data = read_cpld(cpld, 0x04);
        write_cpld(cpld, 0x07, 0x00); // Disable CS for the Boot FLASH/SRAM
	return data;
}

static int aft_shark_cpld_set_clock(wan_aft_cpld_t *cpld, unsigned char media)
{
	int i = 0;
	write_cpld(cpld, 0x00, 0x00);
	for(i=0;i<400;i++);
	switch(media){
	case WAN_MEDIA_T1:
		write_cpld(cpld, 0x00, 0x03);
		break;

	case WAN_MEDIA_E1:
		write_cpld(cpld, 0x00, 0x01);
		break;
	}
	for(i=0;i<400;i++);
	return 0;
}

static int aft_shark_led_test(wan_aft_cpld_t *cpld)
{
	int i = 0;
	unsigned int j = 0, jj = 0;

	// Configuration LED
	write_cpld(cpld, 0x01, 0xc);

	for(i=0;i<20;i++){
		// Turn ON RED
		write_cpld(cpld, 0x01, 0xd);
		for(j=0;j<100000;j++)	
			for(jj=0;jj<100;jj++);	

		// Turn ON GREEN
		write_cpld(cpld, 0x01, 0xe);
		for(j=0;j<100000;j++)
			for(jj=0;jj<100;jj++);	
	}	
	// Turn OFF RED/GREEN
	write_cpld(cpld, 0x01, 0xf);

	return 0;
}

static int
aft_shark_erase_flash(wan_aft_cpld_t *cpld, int stype, int verify)
{
	unsigned long offset = 0x00;
	unsigned char data = 0x00;
	int sector_no = 0;
	int cnt=0;

	for(sector_no = 0; sector_no < 19; sector_no++){
		if (aft_shark_flash_spec[cpld->flash->id][sector_no].sector_type != stype){
			continue;
		}
		offset = aft_shark_flash_spec[cpld->flash->id][sector_no].start_addr;
		write_flash(cpld, stype, MEMORY_TYPE_FLASH, 0xAAA, 0xAA);
		write_flash(cpld, stype, MEMORY_TYPE_FLASH, 0x555, 0x55);
		write_flash(cpld, stype, MEMORY_TYPE_FLASH, 0xAAA, 0x80);
		write_flash(cpld, stype, MEMORY_TYPE_FLASH, 0xAAA, 0xAA);
		write_flash(cpld, stype, MEMORY_TYPE_FLASH, 0x555, 0x55);
		write_flash(cpld, stype, MEMORY_TYPE_FLASH, offset, 0x30);

		do{
//MF			for(i=0;i<100000;i++);
			data = aft_shark_read_flash_byte(
					cpld, stype,
					MEMORY_TYPE_FLASH,
					offset);
			if (data & 0x80){
				break;
			}else if (data & 0x20){
				data = aft_shark_read_flash_byte(
						cpld, stype, 
						MEMORY_TYPE_FLASH, 
						offset);
				if (data & 0x80){
					break;
				}else{
					printf("%s: Failed!\n", __FUNCTION__);
					printf("%s: Sector=%d!\n",
						__FUNCTION__,
						(stype == USER_SECTOR_FLASH) ? 
								sector_no+4:sector_no);
					return -EINVAL;
				}
			}
		} while(1);

		progress_bar("\tErasing sectors\t\t\t");
	}
	printf("\r\tErasing sectors\t\t\tPassed\n");
	if (!verify) return 0;
	
	// Verify that flash is 0xFF
	offset = 0x00;
	if (stype == USER_SECTOR_FLASH){
		offset = AFT_SHARK_USER_SECTOR_START_ADDR;
	}
//        stype = DEF_SECTOR_FLASH;    // M.F debug
//        offset = 0;  

	for(cnt = 0; cnt < 0x80000; cnt++){
//MF		for(i=0;i<10000;i++);
		data = aft_shark_read_flash_byte(cpld, stype, MEMORY_TYPE_FLASH, offset+cnt);	
		if (data != 0xFF){
			printf(" Failed to compare! %05lx -> %02x \n",
						offset+cnt,data);
			return -EINVAL;
		}
		if ((cnt & 0x1FFF) == 0x1000){
			progress_bar("\tErasing sectors (verification)\t");
		}
	}
	printf("\r\tErasing sectors (verification)\tPassed\n");
	return 0;
}

static int
aft_shark_prg_flash_byte(wan_aft_cpld_t *cpld, int stype, unsigned long off32, unsigned char data)
{
	unsigned char data1 = 0x00;

	if (stype == USER_SECTOR_FLASH){
		off32 += AFT_SHARK_USER_SECTOR_START_ADDR;
	}
	write_flash(cpld, stype, MEMORY_TYPE_FLASH, 0xAAA, 0xAA);
	write_flash(cpld, stype, MEMORY_TYPE_FLASH, 0x555, 0x55);
	write_flash(cpld, stype, MEMORY_TYPE_FLASH, 0xAAA, 0xA0);
	write_flash(cpld, stype, MEMORY_TYPE_FLASH, off32, data);

	do{
//MF		for(i=0;i<1000;i++);
		data1 = aft_shark_read_flash_byte(cpld, stype, MEMORY_TYPE_FLASH, off32);
		if ((data1 & 0x80) == (data & 0x80)){
			break;
		}else if (data1 & 0x20){
			data1 = aft_shark_read_flash_byte(
					cpld, stype, 
					MEMORY_TYPE_FLASH, 
					off32);
			if ((data1 & 0x80) == (data & 0x80)){
				break;
			}else{
				printf("prg_flash_byte: Failed (%ld)!\n",
							off32);
				return -EINVAL;
			}
		}
	} while(1);
	return 0;
}

static int aft_shark_reset_flash(wan_aft_cpld_t *cpld)
{
	write_flash(cpld, DEF_SECTOR_FLASH, MEMORY_TYPE_FLASH, 0x00, 0xF0);
	return 0;
}

static int aft_shark_is_protected(wan_aft_cpld_t *cpld, int stype)
{
	return 0;
}

static int aft_shark_flash_id(wan_aft_cpld_t *cpld, int mtype, int stype, int *flash_id)
{
	unsigned char	man_code, device_code;

	aft_shark_reset_flash(cpld);
	write_flash(cpld, stype, mtype, 0xAAA, 0xAA);
	write_flash(cpld, stype, mtype, 0x555, 0x55);
	write_flash(cpld, stype, mtype, 0xAAA, 0x90);

	man_code = aft_shark_read_flash_byte(cpld, stype, mtype, 0x00);
	if (man_code != MCODE_ST){
		printf("The current flash is not supported (man id %02X)!\n",
				man_code);
		return -EINVAL;
	}
	*flash_id = man_code << 8;

	device_code = aft_shark_read_flash_byte(cpld, stype, mtype, 0x02);
	switch(device_code){
	case DCODE_M29W800DT:
		cpld->flash->id = M29W800DT_FID;
		break;
	case DCODE_M29W800DB:
		cpld->flash->id = M29W800DB_FID;
		break;
	default:
		printf("The current flash is not supported (dev id %02X)!\n",
					device_code);
		return -EINVAL;
		break;
	}
	*flash_id |= device_code;
	aft_shark_reset_flash(cpld);
	return 0;
}

static int aft_shark_reload_flash(wan_aft_cpld_t *cpld, int sector_type)
{
	if (sector_type == USER_SECTOR_FLASH){
		/* Reload new code in to Xilinx from
                ** Flash User sector */  
		write_cpld(cpld, 0x07,0xC0); 
	}else{
		/* Reload new code in to Xilinx from
		** Flash Default sector */   
		write_cpld(cpld, 0x07,0x80);   
	}
	return 0;
}

